import asyncio
import logging
import sys
import tempfile
from pathlib import Path

import asyncpg
from faker import Faker
from piou import Cli, Derived, Option

from padmy.anonymize import anonymize_db
from padmy.config import Config
from padmy.db import pretty_print_stats, pprint_compared_dbs, Database
from padmy.logs import setup_logging, logs
from padmy.migration import migration
from padmy.sampling import sample_database, copy_database
from padmy.utils import get_pg_root, get_pg_root_from, init_connection
from padmy.env import CONSOLE

cli = Cli("Padmy utility commands")

cli.add_option("-v", "--verbose", help="Verbosity")
cli.add_option("-vv", "--verbose2", help="Increased verbosity")

cli.add_command_group(migration)


def on_process(verbose: bool = False, verbose2: bool = False):
    setup_logging(logging.DEBUG if verbose2 else logging.INFO if verbose else logging.WARNING)


cli.set_options_processor(on_process)


async def get_explored_db(pg_url: str, db_name: str, schemas: list[str]):
    db = Database(name=db_name)
    logs.info(f"Gathering information about {db_name!r}...")
    async with asyncpg.create_pool(f"{pg_url}/{db_name}", init=init_connection) as pool:
        await db.explore(pool, schemas)
    return db


@cli.command(cmd="anonymize", help="Run anonymization")
async def ano_main(
    pg_url: str = Derived(get_pg_root),
    db_name: str = Option(..., "--db", help="Database to anonymize"),
    config_path: Path = Option(..., "-f", "--file", help="Path to the configuration file"),
):
    config = Config.load_from_file(config_path)
    faker = Faker()
    async with asyncpg.create_pool(f"{pg_url}/{db_name}") as pool:
        await anonymize_db(pool, config, faker)


@cli.command(cmd="sample", help="Sample database")
async def sample_main(
    from_db: str = Option(..., "--db", help="Database to migrate from"),
    to_db: str = Option(..., "--db-to", help="Database to migrate to"),
    schemas: list[str] | None = Option(None, "--schemas", help="Schemas to samples"),
    sample: int | None = Option(None, "--sample", help="Sample size"),
    config_path: Path | None = Option(None, "-f", "--file", help="Path to the configuration file"),
    copy_db: bool = Option(False, "--copy-db", help="Copy database before sampling"),
    # pg_url: str = Derived(get_pg_root),
    pg_from: str = Derived(get_pg_root_from("from")),
    pg_to: str = Derived(get_pg_root_from("to")),
    progress: bool = Option(False, "--progress", help="Show sampling progress"),
):
    """
    Create a copy of a given database and inserts a sample to a new database
    """

    if config_path:
        config = Config.load_from_file(config_path)
    elif sample is not None and schemas is not None:
        config = Config.load(sample, schemas=schemas)
    else:
        raise NotImplementedError("Please specify either a sample size and schemas or a config path")

    _schemas = [x.schema for x in config.schemas]
    if copy_db:
        logs.info(f"Copying {from_db!r} to {to_db!r}")
        try:
            copy_database(
                from_pg_uri=pg_from,
                to_pg_uri=pg_to,
                from_db=from_db,
                to_db=to_db,
                schemas=_schemas,
            )
        except Exception as e:
            logs.error(e)
            sys.exit(1)
        else:
            logs.info("Done!")

    # TODO clean if error
    conn = await asyncpg.connect(f"{pg_from}/{from_db}", statement_cache_size=0)
    target_conn = await asyncpg.connect(f"{pg_to}/{to_db}")

    db = await get_explored_db(pg_from, from_db, _schemas)
    db.load_config(config)

    pretty_print_stats(db)

    try:
        await sample_database(conn, target_conn, db, show_progress=progress)
    except Exception as e:
        raise e
    finally:
        await asyncio.wait_for(conn.close(), timeout=1)
        await asyncio.wait_for(target_conn.close(), timeout=1)

    new_db = await get_explored_db(pg_to, to_db, _schemas)

    pprint_compared_dbs(db, new_db)


@cli.command(cmd="copy-db", help="Copy a database schema to a new database")
def copy_main(
    from_db: str = Option(..., "--db", help="Database to migrate from"),
    to_db: str = Option(..., "--db-to", help="Database to migrate to"),
    schemas: list[str] = Option(..., "--schemas", help="Schemas to samples"),
    pg_from: str = Derived(get_pg_root_from("from")),
    pg_to: str = Derived(get_pg_root_from("to")),
):
    from padmy.sampling import copy_database

    logs.info(f"Copying {from_db!r} to {to_db!r}")
    copy_database(
        from_pg_uri=pg_from,
        to_pg_uri=pg_to,
        from_db=from_db,
        to_db=to_db,
        schemas=schemas,
        drop_public=False,
    )
    logs.info("Done!")


@cli.command(cmd="analyze", help="Analyze database")
async def analyze_main(
    db_name: str = Option(..., "--db", help="Database to explore"),
    pg_url: str = Derived(get_pg_root),
    schemas: list[str] = Option(..., "--schemas", help="Schemas to analyze"),
    show_graphs: bool = Option(False, "--show-graphs", help="Show graphs of the database"),
    port: int = Option(5555, "--graph-port", help="Port for the graph"),
):
    """
    Show information about a database
    """
    from padmy.db import pretty_print_stats

    db = await get_explored_db(pg_url, db_name, schemas)

    if not show_graphs:
        pretty_print_stats(db)
        return

    from padmy.sampling.viz import run_simple_app

    run_simple_app(db, port=port)


@cli.command(cmd="compare", help="Compare 2 databases")
async def compare_db_main(
    from_db: str = Option(..., "--db", help="Database to explore"),
    to_db: str = Option(..., "--db-to", help="Database to explore"),
    pg_url: str = Derived(get_pg_root),
    schemas: list[str] = Option(..., "--schemas", help="Schemas to analyze"),
):
    from padmy.db import pprint_compared_dbs

    db1, db2 = await asyncio.gather(
        get_explored_db(pg_url, from_db, schemas),
        get_explored_db(pg_url, to_db, schemas),
    )

    pprint_compared_dbs(db1, db2)


@cli.command(cmd="schema-diff", help="Compare the schemas of 2 databases")
def schema_diff(
    pg_from: str = Derived(get_pg_root_from("from")),
    pg_to: str = Derived(get_pg_root_from("to")),
    database: str = Option(..., "--db", help="Database to compare"),
    schemas: list[str] = Option(..., "--schemas", help="Schemas to analyze"),
    no_privileges: bool = Option(False, "-x", "--no-privileges", help="Exclude privileges from the dump"),
):
    from padmy.compare import compare_databases

    with tempfile.TemporaryDirectory() as dump_dir:
        _dump_dir = Path(dump_dir)
        diff = compare_databases(
            pg_from,
            to_pg_url=pg_to,
            schemas=schemas,
            database=database,
            dump_dir=_dump_dir,
            no_privileges=no_privileges,
        )
    if diff is not None:
        CONSOLE.print("[orange]Differences found[/orange]")
        CONSOLE.print("\n".join(diff))
    else:
        CONSOLE.print("[green]No differences found[/green]")


def run():
    cli.run()


if __name__ == "__main__":
    run()



from padmy.logs import logs

try:
    import networkx as nx
except ImportError:
    logs.error("Please install networkx to use this module")
    raise

from ..db import Database, Table


def add_nodes(graph: nx.Graph, table: Table, processed_nodes: set[str]):
    if table.full_name in processed_nodes:
        return

    graph.add_node(table.full_name.replace(".", "_"), count=table.count, label=table.full_name)

    processed_nodes.add(table.full_name)

    for _child_table in table.child_tables:
        graph.add_edge(
            table.full_name.replace(".", "_"),
            _child_table.full_name.replace(".", "_"),
        )
        add_nodes(graph, _child_table, processed_nodes)


def convert_db(db: Database) -> nx.DiGraph:
    g = nx.DiGraph()
    processed_nodes = set()

    for table in db.tables:
        add_nodes(g, table, processed_nodes)

    return g



import json
from dataclasses import dataclass, asdict
from padmy.env import PADMY_CONFIG
from .utils import get_git_email


@dataclass
class MigrationConfig:
    author: str

    @property
    def data(self):
        return asdict(self)

    @classmethod
    def load(cls):
        _data = json.loads(PADMY_CONFIG.read_text()) if PADMY_CONFIG.exists() else {}
        return cls(author=_data.get("author") or get_git_email())

    def save(self):
        if not PADMY_CONFIG.parent.exists():
            PADMY_CONFIG.parent.mkdir(parents=True)
        with PADMY_CONFIG.open("w") as f:
            json.dump(self.data, f, indent=4)



import asyncio
import functools
from dataclasses import dataclass, field

import asyncpg
from rich.table import Table as RTable

from padmy.config import Config, ConfigTable, ConfigSchema
from padmy.logs import logs
from padmy.utils import get_first, get_conn
from padmy.env import CONSOLE
import sys

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


def _get_full_name(schema: str | None, table: str | None) -> str:
    if schema is None or table is None:
        raise ValueError("schema and table must not be empty")
    return f"{schema}.{table}"


@dataclass
class FKConstraint:
    column_names: list[str]
    constraint_name: str

    # references
    foreign_schema: str
    foreign_table: str
    foreign_column_names: list[str]

    table: str | None = None
    schema: str | None = None

    @property
    def foreign_full_name(self):
        return _get_full_name(self.foreign_schema, self.foreign_table)

    @property
    def full_name(self):
        return _get_full_name(self.schema, self.table)


@dataclass
class PKConstraint:
    column_name: str
    table: str
    schema: str

    @property
    def full_name(self):
        return _get_full_name(self.schema, self.table)


@dataclass
class Column:
    name: str
    is_generated: bool


@dataclass(eq=False)
class Table:
    schema: str
    table: str

    columns: list[Column] | None = None

    _count: int | None = None  # field(init=False)

    foreign_keys: list[FKConstraint] = field(default_factory=list)
    primary_keys: list[PKConstraint] = field(default_factory=list)

    parent_tables: set[Self] = field(default_factory=set)
    child_tables: set[Self] = field(default_factory=set)

    # Has already been sampled, and it's temporary table has been created
    has_been_processed: bool = False

    # Sample size
    sample_size: int | None = None
    # Ignore this table
    ignore: bool = False

    @property
    def parent_tables_safe(self):
        # return self.parent_tables - {self}
        return {x for x in self.parent_tables if x.full_name != self.full_name and not x.ignore}

    @property
    def child_tables_safe(self):
        # Returns the child table that are not the current table
        return {x for x in self.child_tables if x.full_name != self.full_name and not x.ignore}

    @property
    def full_name(self):
        return _get_full_name(self.schema, self.table)

    @property
    def tmp_name(self):
        return f"_{self.schema}_{self.table}_tmp"

    @property
    def has_parent(self):
        return len(self.parent_tables_safe) > 0

    @property
    def has_children(self):
        return len(self.child_tables_safe) > 0

    @property
    def children_has_been_processed(self):
        return all(_child.has_been_processed for _child in self.child_tables_safe)

    @property
    def count(self) -> int:
        if self._count is None:
            raise ValueError("Count must be loaded first")
        return self._count

    @count.setter
    def count(self, v: int):
        self._count = v

    def get_values(self, table: str | None = None):
        if self.columns is None:
            raise ValueError("Columns must be loaded first")
        _table = f"{table}." if table is not None else ""
        return ", ".join(sorted([f'{_table}"{x.name}"' for x in self.columns if not x.is_generated]))

    @property
    def values(self):
        return self.get_values()

    @property
    def is_leaf(self):
        return not self.has_children and self.has_parent

    @property
    def is_root(self):
        return not self.has_parent

    async def load_count(self, conn: asyncpg.Connection):
        self._count = await conn.fetchval(f"SELECT count(*) from {self.full_name}")

    def __eq__(self, other: object):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return hash((self.full_name, self.has_been_processed))

    def __repr__(self):
        return (
            f"Table(full_name={self.full_name!r} "
            f"count={self._count} "
            f"foreign_keys={len(self.foreign_keys)} "
            f"parents={len(self.parent_tables)} "
            f"children={len(self.child_tables)}"
            f")"
        )

    def __post_init__(self):
        for fk in self.foreign_keys:
            fk.table = self.table
            fk.schema = self.schema


async def get_tables(conn: asyncpg.Connection, schemas: list[str]):
    query = """
    SELECT 
    table_schema AS schema, 
    table_name AS table
    FROM information_schema.tables
    WHERE table_schema = ANY ($1::TEXT[]) AND
    table_type = 'BASE TABLE'
    ORDER BY table_schema, table_name
    """
    data = await conn.fetch(query, schemas)
    return [Table(**x) for x in data]


async def get_columns(conn: asyncpg.Connection, tables: list[Table]) -> dict[str, list[Column]]:
    query = """
    SELECT full_name,
           JSON_AGG(
                   JSON_BUILD_OBJECT(
                           'name', column_name,
                           'is_generated', COALESCE(generation_expression IS NOT NULL OR identity_generation = 'ALWAYS', FALSE)
                       )
               ) AS columns
    FROM (SELECT *, table_schema || '.' || table_name AS full_name FROM information_schema.columns) t
    WHERE full_name = ANY ($1::TEXT[])
    GROUP BY full_name
    """
    data = await conn.fetch(query, [x.full_name for x in tables])
    columns = {}
    for x in data:
        print(x["columns"])
        columns[x["full_name"]] = [Column(**col) for col in x["columns"]]
    return columns


SCHEMA_FK_QUERY = """
SELECT c.conname                                         AS constraint_name,
       sch.nspname                                       AS schema_name,
       tbl.relname                                       AS table_name,
       ARRAY_AGG(col.attname ORDER BY u.attposition)     AS column_names,
       f_sch.nspname                                     AS foreign_schema_name,
       f_tbl.relname                                     AS foreign_table_name,
       ARRAY_AGG(f_col.attname ORDER BY f_u.attposition) AS foreign_column_names
FROM pg_constraint c
         LEFT JOIN LATERAL UNNEST(c.conkey) WITH ORDINALITY AS u(attnum, attposition) ON TRUE
         LEFT JOIN LATERAL UNNEST(c.confkey) WITH ORDINALITY AS f_u(attnum, attposition)
                   ON f_u.attposition = u.attposition
         JOIN pg_class tbl ON tbl.oid = c.conrelid
         JOIN pg_namespace sch ON sch.oid = tbl.relnamespace
         LEFT JOIN pg_attribute col ON (col.attrelid = tbl.oid AND col.attnum = u.attnum)
         LEFT JOIN pg_class f_tbl ON f_tbl.oid = c.confrelid
         LEFT JOIN pg_namespace f_sch ON f_sch.oid = f_tbl.relnamespace
         LEFT JOIN pg_attribute f_col ON (f_col.attrelid = f_tbl.oid AND f_col.attnum = f_u.attnum)
WHERE c.contype = 'f' 
    AND sch.nspname = ANY ($1::TEXT[]) 
GROUP BY constraint_name, "schema_name", "table_name", f_sch.nspname, f_tbl.relname
ORDER BY "schema_name", "table_name", "constraint_name";
"""

SCHEMA_PK_QUERY = """
SELECT 
    tc.table_schema AS schema, 
    tc.table_name AS table, 
    c.column_name
FROM information_schema.table_constraints tc
JOIN information_schema.constraint_column_usage AS ccu USING (constraint_schema, constraint_name)
JOIN information_schema.columns AS c ON c.table_schema = tc.constraint_schema
  AND tc.table_name = c.table_name AND ccu.column_name = c.column_name
WHERE constraint_type = 'PRIMARY KEY' AND tc.table_schema = ANY ($1::TEXT[]);
"""


async def load_foreign_keys(conn: asyncpg.Connection, schemas: list[str]):
    data = await conn.fetch(SCHEMA_FK_QUERY, schemas)
    return [
        FKConstraint(
            column_names=x["column_names"],
            constraint_name=x["constraint_name"],
            foreign_schema=x["foreign_schema_name"],
            foreign_table=x["foreign_table_name"],
            foreign_column_names=x["foreign_column_names"],
            table=x["table_name"],
            schema=x["schema_name"],
        )
        for x in data
    ]


async def load_primary_keys(conn: asyncpg.Connection, schemas: list[str]):
    data = await conn.fetch(SCHEMA_PK_QUERY, schemas)
    return [PKConstraint(**x) for x in data]


GET_COLUMNS_TYPE_QUERY = """
SELECT 
    column_name, data_type 
FROM information_schema.columns 
WHERE table_schema = $1 AND 
      table_name = $2 AND 
      column_name = ANY ($3::TEXT[])
"""


async def load_columns_type(conn: asyncpg.Connection, schema: str, table: str, columns: list[str]):
    data = await conn.fetch(GET_COLUMNS_TYPE_QUERY, schema, table, columns)
    return functools.reduce(lambda p, n: {**p, **{n["column_name"]: n["data_type"]}}, data, {})


@dataclass
class Database:
    name: str
    tables: list[Table] = field(default_factory=list)

    async def explore(
        self,
        pool: asyncpg.Pool,
        schemas: list[str] | None = None,
        *,
        load_count: bool = True,
    ):
        _schemas = schemas or list(set(x.schema for x in self.tables))
        if not _schemas:
            raise ValueError("No schemas to explore")

        _schemas_str = ", ".format()
        async with pool.acquire() as conn:
            logs.debug(f"Loading tables for {_schemas_str}...")
            self.tables = await get_tables(conn, _schemas)
            logs.debug(f"{len(self.tables)} tables found")

            logs.debug("Loading table columns...")
            columns = await get_columns(conn, self.tables)
            for table in self.tables:
                table.columns = columns[table.full_name]
            logs.debug("Loading foreign keys...")
            fks = await load_foreign_keys(conn, _schemas)
            logs.debug(f"{len(fks)} tables found")
            logs.debug("Loading primary keys...")
            pks = await load_primary_keys(conn, _schemas)
            logs.debug(f"{len(pks)} primary keys found")

        _tables: dict[str, Table] = {_table.full_name: _table for _table in self.tables}

        for _pk in pks:
            _tables[_pk.full_name].primary_keys.append(_pk)

        for _fk in fks:
            _tables[_fk.full_name].foreign_keys.append(_fk)
            _tables[_fk.full_name].parent_tables.add(_tables[_fk.foreign_full_name])
            _tables[_fk.foreign_full_name].child_tables.add(_tables[_fk.full_name])

        if load_count:
            logs.info("Counting number of rows in tables...")
            await asyncio.gather(*[get_conn(pool, table.load_count) for table in self.tables])

    def load_config(self, config: Config):
        """
        Loads the sample sizes for each tables from the config file.
        Tables need to have been loaded first
        """
        if not self.tables:
            raise ValueError("Tables must be loaded first")

        _schemas: dict[str, ConfigSchema] = {schema.schema: schema for schema in config.schemas}
        _tables: dict[str, ConfigTable] = {f"{_table.schema}.{_table.table}": _table for _table in config.tables}
        for _table in self.tables:
            _config_schema = _schemas.get(_table.schema)
            _config_table = _tables.get(_table.full_name)
            _sample = get_first(
                _config_table.sample if _config_table is not None else None,
                _config_schema.sample if _config_schema is not None else None,
                config.sample,
                fn=lambda x: x is not None,
            )
            if _sample is None:
                raise ValueError("_sample must not be empty")
            _table.sample_size = int(_sample)
            if _config_table is not None:
                _table.ignore = _config_table.ignore


def pretty_print_stats(database: Database):
    table = RTable(title=f"Stats for {database.name}")

    table.add_column("Table", justify="left", style="cyan")
    table.add_column("Count", justify="right", style="green")
    table.add_column("# FKs", justify="right", style="magenta")
    table.add_column("# Parents", justify="right", style="magenta")
    table.add_column("# Children", justify="right", style="magenta")

    for _table in sorted(database.tables, key=lambda x: x.full_name):
        table.add_row(
            _table.full_name,
            str(_table.count),
            str(len(_table.foreign_keys)),
            str(len(_table.parent_tables)),
            str(len(_table.child_tables)),
        )

    CONSOLE.print(table)


def pprint_compared_dbs(db_1: Database, db_2: Database):
    table = RTable(title=f"Comparing {db_1.name} and {db_2.name}")

    table.add_column("Table", justify="left", style="blue")
    table.add_column(f"Count {db_1.name!r}", justify="right", style="cyan")
    table.add_column(f"Count {db_2.name!r}", justify="right", style="cyan")
    table.add_column("Diff", justify="right", style="green")

    tables_1, tables_2 = sorted(db_1.tables, key=lambda x: x.full_name), sorted(db_2.tables, key=lambda x: x.full_name)

    for _table1, _table2 in zip(tables_1, tables_2):
        perc_diff = 100 if _table2.count == 0 else int(_table2.count * 100 / _table1.count)
        if perc_diff > 0:
            pass
        elif perc_diff < 0:
            pass
        else:
            pass
        table.add_row(_table1.full_name, str(_table1.count), str(_table2.count), f"{perc_diff}%")

    CONSOLE.print(table)



import dataclasses
import datetime as dt
import difflib
import filecmp
import functools
import textwrap
import typing
from contextlib import nullcontext
from pathlib import Path
from typing import Callable, Iterator

import asyncpg
from asyncpg import Connection
from asyncpg.exceptions import UndefinedTableError
from tracktolib.pg import insert_one
from tracktolib.utils import get_chunks

from padmy.logs import logs
from padmy.utils import exec_file, pg_dump, exec_psql_file, remove_restrict_clauses
from .utils import get_files, iter_migration_files, MigrationFile

_GET_LATEST_MIGRATION_QUERY = textwrap.dedent(
    """
SELECT _m.file_ts, _m.file_name, _m.file_id
FROM public.migration _m
       left join public.migration _r
                 on _r.migration_type = 'down' and _r.file_id = _m.file_id
where _m.migration_type = 'up'
and _r.file_id is null
order by _m.applied_at desc, file_ts desc limit 1
"""
)

# PG_GLOBAL_DUMP_DIR = GLOBAL_TMP / 'matching-pg-backup' / 'dump'
SETUP_FILE = Path(__file__).parent / "db.sql"


class MigrationError(Exception):
    def __init__(self, msg: str, diff: str):
        self.msg = msg
        self.diff = diff


async def migrate_setup(conn: Connection):
    logs.info("Setting up migration tables")
    await exec_file(conn, SETUP_FILE)
    logs.info("Done")


CompareFilesFn = Callable[[Path, Path], Iterator[str] | None]


def compare_files(file1: Path, file2: Path, ignore_restrict: bool = True) -> Iterator[str] | None:
    """
    Compares 2 files. If they are the same, returns None, otherwise returns the diff.
    If ignore_restrict is True, ignores \restrict clauses in the
    diff ([PostgreSQL 17.6+](https://www.postgresql.org/docs/17/release-17-6.html#RELEASE-17-6-CHANGES) feature).
    """
    if ignore_restrict:
        remove_restrict_clauses(file1)
        remove_restrict_clauses(file2)
    _no_diff = filecmp.cmp(file1, file2)
    if _no_diff:
        return None
    diff = difflib.unified_diff(
        file1.read_text().split("\n"),
        file2.read_text().split("\n"),
        fromfile=file1.name,
        tofile=file2.name,
    )
    return diff


default_compare_files: CompareFilesFn = functools.partial(compare_files, ignore_restrict=True)


def _verify_migration(
    database: str,
    schemas: list[str],
    migration_id: str,
    up_file: Path,
    down_file: Path,
    dump_dir: Path,
    *,
    compare_files_fn: CompareFilesFn = default_compare_files,
):
    _before_dump, _after_dump = (
        f"{migration_id}-before.sql",
        f"{migration_id}-after.sql",
    )

    # Dump before
    pg_dump(
        database,
        schemas,
        dump_path=str(dump_dir / _before_dump),
        options=["-E", "utf8", "--schema-only"],
    )

    logs.info(f"Applying {up_file.name}")
    exec_psql_file(database, str(up_file))

    logs.info(f"Applying {down_file.name}")
    exec_psql_file(database, str(down_file))

    # Dump after
    pg_dump(
        database,
        schemas,
        dump_path=str(dump_dir / _after_dump),
        options=["-E", "utf8", "--schema-only"],
    )

    _before_dump_file, _after_dump_file = (
        dump_dir / _before_dump,
        dump_dir / _after_dump,
    )

    _diff = compare_files_fn(_before_dump_file, _after_dump_file)
    if _diff is not None:
        raise MigrationError(f"Difference found for migration: {migration_id}", diff="\n".join(_diff))


def migrate_verify(
    database: str,
    schemas: list[str],
    dump_dir: Path,
    migration_folder: Path,
    *,
    only_last: bool = False,
    compare_files_fn: CompareFilesFn = default_compare_files,
    skip_down_restore: bool = False,
):
    """
    Verifies that the up/down migration is correct
    """
    _pg_dump = functools.partial(pg_dump, database=database, schemas=schemas)

    migration_files = get_files(migration_folder)
    logs.info(f"Found {len(migration_files)} migration files")

    logs.info("Checking...")
    _down_files: list[MigrationFile] = []
    _nb_migrations = int(len(migration_files) / 2)
    try:
        for i, (_up_file, _down_file) in enumerate(iter_migration_files(migration_files)):
            _skip = (
                _down_file.skip_verify
                or
                # We verify all if only_last is False or if we only have one migration (up/down)
                (only_last and i != _nb_migrations - 1 and len(migration_files) != 2)
            )
            if not _skip:
                logs.info(f"Checking migration {i + 1}/{_nb_migrations}")
                _verify_migration(
                    database=database,
                    schemas=schemas,
                    migration_id=_up_file.file_id,
                    up_file=_up_file.path,
                    down_file=_down_file.path,
                    dump_dir=dump_dir,
                    compare_files_fn=compare_files_fn,
                )
            else:
                logs.info(f"Skipping migration {_up_file.file_id}")
            # Need to be reapplied for the next tests
            exec_psql_file(database, str(_up_file.path))
            _down_files.append(_down_file)
    except MigrationError as e:
        logs.error(e.msg)
        logs.error(e.diff)
        raise e

    if not skip_down_restore:
        # Restoring to initial state
        logs.info("Restoring to initial state")
        for i, _down_file in enumerate(reversed(_down_files)):
            logs.info(f"Applying {_down_file.path.name} ({len(_down_files) - i}/{len(_down_files)})")
            exec_psql_file(database, str(_down_file.path))
        # else:
        #     if last_down_file:
        #         exec_psql_file(database, str(last_down_file.path))

    logs.info("Everything ok!")


@dataclasses.dataclass
class LatestMigration:
    timestamp: dt.datetime
    file_name: str
    file_id: str

    @classmethod
    def load(cls, data: dict):
        return cls(timestamp=data["file_ts"], file_name=data["file_name"], file_id=data["file_id"])


class NoSetupTableError(Exception):
    pass


async def _get_latest_migration(conn: Connection):
    try:
        resp = await conn.fetchrow(_GET_LATEST_MIGRATION_QUERY)
    except UndefinedTableError as e:
        if 'relation "public.migration" does not exist' in str(e):
            raise NoSetupTableError(
                'Could not find table "public.migration", did you forget to setup the table'
                ' by running "migration setup" ?'
            )
        else:
            raise
    return LatestMigration.load(dict(resp)) if resp else None


async def get_migration_files(
    conn: asyncpg.Connection, folder: Path, *, nb_migrations: int = -1
) -> list[MigrationFile]:
    latest_migration = await _get_latest_migration(conn)
    logs.info(
        f"Latest timestamp: {latest_migration.timestamp} ({latest_migration.file_id})"
        if latest_migration
        else "No previous migration found"
    )
    migration_files = get_files(folder)

    migrations_to_apply = [
        _up_file
        for _up_file, _ in iter_migration_files(migration_files)
        if (latest_migration is None)
        or (_up_file.ts >= latest_migration.timestamp and _up_file.path.name != latest_migration.file_name)
    ]

    if nb_migrations > 0:
        migrations_to_apply = migrations_to_apply[:nb_migrations]
    return migrations_to_apply


async def apply_migration(conn: asyncpg.Connection, migration: MigrationFile, metadata: dict | None = None):
    logs.info(f"Running {migration.path.name} (ts: {migration.ts})...")
    try:
        await exec_file(conn, migration.path)
    except Exception as e:
        logs.error(f'Failed to execute migration_id "{migration.file_id}" ({e})')
        raise e
    migration_data = {
        "file_ts": migration.ts,
        "file_id": migration.file_id,
        "migration_type": "up",
        "file_name": migration.path.name,
    }
    if metadata:
        migration_data["meta"] = metadata
    await insert_one(conn, "public.migration", migration_data)


async def migrate_up(
    conn: Connection,
    folder: Path,
    *,
    nb_migrations: int = -1,
    metadata: dict | None = None,
    use_transaction: bool = True,
) -> bool:
    """Migrate from the latest migration applied available in the database
    to the latest one in the `folder` dir.
    """
    migration_files = await get_migration_files(conn, folder, nb_migrations=nb_migrations)

    if not migration_files:
        logs.info("No migrations to apply")
        return False

    _transaction = conn.transaction if use_transaction else nullcontext
    logs.info(f"Found {len(migration_files)} migrations to apply")
    async with _transaction():
        for _migration in migration_files:
            await apply_migration(conn, _migration, metadata)
    logs.info("Done!")
    return True


class AppliedMigration(typing.TypedDict):
    file_id: str
    has_applied_rollback: bool


async def get_applied_migrations(conn: asyncpg.Connection) -> list[AppliedMigration]:
    try:
        data = await conn.fetch(
            """
            SELECT _m.file_id, _r.file_id is not null as has_applied_rollback
            FROM public.migration _m
                     left join public.migration _r on _r.migration_type = 'down' and _r.file_id = _m.file_id
            where _m.migration_type = 'up'
            order by _m.applied_at desc
            """
        )
    except UndefinedTableError as e:
        if 'relation "public.migration" does not exist' in str(e):
            raise NoSetupTableError(
                'Could not find table "public.migration", did you forget to setup the table'
                ' by running "migration setup" ?'
            )
        else:
            raise
    return [
        AppliedMigration(file_id=item["file_id"], has_applied_rollback=item["has_applied_rollback"]) for item in data
    ]


async def get_rollback_files(
    conn: asyncpg.Connection,
    folder: Path,
    *,
    nb_migrations: int = -1,
    migration_id: str | None = None,
) -> list[MigrationFile]:
    if migration_id is not None and nb_migrations >= 0:
        raise ValueError('Specify either "nb_migrations" or "migration_id"')

    applied_migrations = await get_applied_migrations(conn)
    applied_migration_ids = {item["file_id"] for item in applied_migrations if not item["has_applied_rollback"]}
    rollback_to_apply = [
        _down_file
        for _up_file, _down_file in iter_migration_files(get_files(folder))
        if _down_file.file_id in applied_migration_ids
    ][::-1]

    if migration_id is not None:
        for idx, _rollback in enumerate(rollback_to_apply):
            if _rollback.file_id == migration_id:
                rollback_to_apply = rollback_to_apply[: idx + 1]
                break
        else:
            raise ValueError(f"Could not find migration_id {migration_id!r}")
    elif nb_migrations > 0:
        rollback_to_apply = rollback_to_apply[:nb_migrations]

    return rollback_to_apply


async def migrate_down(
    conn: Connection,
    folder: Path,
    *,
    nb_migrations: int | None = None,
    metadata: dict | None = None,
    use_transaction: bool = True,
    migration_id: str | None = None,
) -> bool:
    """
    Rollback `nb_migrations` back (default -1 for all available) or until migration_id (if set)
    """
    rollback_files = await get_rollback_files(
        conn, folder, migration_id=migration_id, nb_migrations=nb_migrations or -1
    )
    nb_files = len(rollback_files)
    if nb_files == 0:
        logs.info("No rollback files to apply")
        return False

    _transaction = conn.transaction if use_transaction else nullcontext
    logs.info(f"Found {nb_files} rollback files to apply")
    async with _transaction():
        for _rollback in rollback_files:
            logs.info(f"Running {_rollback.path.name}...")
            try:
                await exec_file(conn, _rollback.path)
            except Exception as e:
                logs.error(f'Failed to execute migration_id "{_rollback.file_id}"')
                raise e

            migration_data = {
                "file_ts": _rollback.ts,
                "file_id": _rollback.file_id,
                "migration_type": "down",
                "file_name": _rollback.path.name,
            }
            if metadata:
                migration_data["meta"] = metadata
            await insert_one(conn, "public.migration", migration_data)

    logs.info("Done!")
    return True


async def get_missing_migrations(conn: asyncpg.Connection, folder: Path, *, chunk_size: int = 500):
    files = get_files(folder, up_only=True)
    not_applied_files = []
    for chunk in get_chunks(files, size=chunk_size, as_list=True):
        file_ids = await conn.fetch(
            """
            SELECT file_id
            from migration
            where migration_type = 'up'
              and file_id = any ($1)
            """,
            [file.file_id for file in chunk],
        )
        file_ids_db = set(file["file_id"] for file in file_ids)
        for _file in chunk:
            if _file.file_id not in file_ids_db:
                not_applied_files.append(_file)

    return not_applied_files


async def verify_migrations(
    conn: asyncpg.Connection, folder: Path, *, chunk_size: int = 500, metadata: dict | None = None
):
    """
    Verify that all migrations inside "folder" have been applied to the database.
    If not, tries to apply them
    """
    not_applied_files = await get_missing_migrations(conn, folder=folder, chunk_size=chunk_size)
    if not not_applied_files:
        logs.info("All migrations have been applied")
        return
    logs.info(f"Found {len(not_applied_files)} missing files")
    _meta = metadata or {"missing": True}
    for _file in not_applied_files:
        await apply_migration(conn, _file, metadata=_meta)



import time
from pathlib import Path

from rich.markup import escape
from rich.prompt import Prompt

from padmy.env import CONSOLE
from padmy.logs import logs
from .migration import MigrationFile
from .config import MigrationConfig
from .utils import get_files, iter_migration_files, Header


def _get_user_email() -> str | None:
    _config = MigrationConfig.load()
    author = Prompt.ask("[blue]Author[/blue]", default=_config.author, console=CONSOLE)
    _config.author = author
    _config.save()
    return author


def _get_last_migration_name(folder: Path) -> str | None:
    """Returns the most recent migration files"""
    files = get_files(reverse=True, folder=folder)
    if not files:
        return None
    up_file, down_file = next(iter_migration_files(files))
    return up_file.path.name


def create_new_migration(
    folder: Path,
    version: str | None = None,
    user_email: str | None = None,
    skip_verify: bool = False,
) -> tuple[Path, Path]:
    """
    Creates 2 new files, up and down
    """
    folder.mkdir(exist_ok=True, parents=True)

    _base_name = MigrationFile.generate_base_name(ts=int(time.time()))
    CONSOLE.print(f"\nCreating new migration file ([green]{escape(_base_name)}[/green])\n")

    last_migration = _get_last_migration_name(folder)
    logs.debug(f"Last migration files: {last_migration}")
    author = user_email or _get_user_email()
    logs.debug(f"User email: {author}")

    up_file = folder / Path(f"{_base_name}-up.sql")
    down_file = folder / Path(f"{_base_name}-down.sql")

    _header = Header(last_migration, author, version)
    up_file.write_text(_header.as_text())
    _header.skip_verify = skip_verify
    down_file.write_text(_header.as_text().replace("-up", "-down"))

    CONSOLE.print("\nNew files created!\n")
    return up_file, down_file



import difflib
import filecmp
from pathlib import Path
from typing import Iterator

from padmy.utils import pg_dump, temp_pg_env, remove_restrict_clauses


def compare_databases(
    from_pg_url: str,
    database: str,
    schemas: list[str],
    dump_dir: Path,
    *,
    to_pg_url: str | None = None,
    db_to: str | None = None,
    no_privileges: bool = True,
    ignore_restrict: bool = True,
) -> Iterator[str] | None:
    """
    Compare the schemas of 2 databases.
    If ignore_restrict is True, ignores \restrict clauses in the
    diff ([PostgreSQL 17.6+](https://www.postgresql.org/docs/17/release-17-6.html#RELEASE-17-6-CHANGES) feature).

    """
    dump_1 = dump_dir / f"{database}-from.sql"
    db_to = db_to or database
    dump_2 = dump_dir / f"{db_to}-to.sql"
    _cmd = ["-E", "utf8", "--schema-only"]
    if no_privileges:
        _cmd.append("--no-privileges")

    with temp_pg_env(from_pg_url):
        pg_dump(database, schemas, dump_path=str(dump_1), options=_cmd, get_env=False)
        if ignore_restrict:
            remove_restrict_clauses(dump_1)

    _to_pg_url = to_pg_url or from_pg_url
    with temp_pg_env(_to_pg_url):
        pg_dump(db_to, schemas, dump_path=str(dump_2), options=_cmd, get_env=False)
        if ignore_restrict:
            remove_restrict_clauses(dump_2)

    _no_diff = filecmp.cmp(dump_1, dump_2)

    if not _no_diff:
        diff = difflib.unified_diff(
            dump_1.read_text().split("\n"),
            dump_2.read_text().split("\n"),
            fromfile=dump_1.name,
            tofile=dump_2.name,
        )
        return diff
    return None



from importlib import metadata

from padmy.run import run

try:
    __version__ = metadata.version("padmy")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0"

__all__ = (
    "__version__",
    "run",
)



import os
import shutil
import tempfile
from pathlib import Path
from typing import Literal

from asyncpg import Connection
from padmy.env import SQL_DIR, MIGRATION_DIR, CONSOLE
from padmy.logs import logs
from padmy.utils import get_pg, PgHost, PgPort, PgUser, PgDatabase, PgPassword
from piou import Option, Derived, CommandGroup, Password, CommandError

MIGRATION_DESCRIPTION = """
Utilities to handle schema migrations with PostgresSQL.  
If it's the first time running theses scripts please run 
**`poetry run cli migrate setup`** to setup the database's table
"""

migration = CommandGroup("migrate", help="Migration utilities", description=MIGRATION_DESCRIPTION)

MigrationDir = Option(MIGRATION_DIR or ..., "--sql-dir", help="Directory containing the migration files")
SQLDir = Option(SQL_DIR or ..., "--sql-dir", help="Directory containing the table definitions")


@migration.command(cmd="new-sql")
def new_sql_file(
    sql_dir: Path = SQLDir,
    position: int = Option(1, help="Position where to insert the sql file"),
):
    """
    Utility to insert a new sql file between other sql files.
    """
    from .new_sql import create_sql_file

    new_file = create_sql_file(sql_dir, position)
    logs.info(f"Created new sql file: {new_file}")


@migration.command(cmd="apply-sql")
async def apply_sql_files(
    pg_conn: Connection = Derived(get_pg),
    sql_dir: Path = SQLDir,
):
    """
    Utility to insert a new sql file between other sql files.
    """
    from ..utils import exec_file

    for file in sorted(sql_dir.glob("*.sql")):
        logs.info(f"Applying {file.name!r}")
        await exec_file(pg_conn, file)


@migration.command(cmd="new", help="Creates 2 new files for a migration (up and down)")
def new_migrate_file(
    migration_folder: Path = MigrationDir,
    version: str | None = Option(None, "--version", help="Version of the migration"),
    user_email: str | None = Option(None, "--author", help="Author of the migration"),
    skip_verify: bool = Option(False, "--skip-verify", help="Should the down file be verified or not"),
):
    """
    Creates 2 new files for a migration (up and down).
    When applying *up.sql* then *down.sql*, the database must be
    in the state it was previously.
    To check if your **down** file works as expected, you can use the `verify` command.
    """
    from .create_files import create_new_migration

    create_new_migration(migration_folder, version=version, user_email=user_email, skip_verify=skip_verify)


@migration.command(cmd="up", help="Migrate database to the new schema")
async def migrate_up_main(
    pg_conn: Connection = Derived(get_pg),
    sql_dir: Path = MigrationDir,
    nb_migrations: int = Option(1, "-n", "--nb-migrations", help="Number of migrations to apply"),
):
    from .migration import migrate_up, NoSetupTableError

    try:
        await migrate_up(pg_conn, folder=sql_dir, nb_migrations=nb_migrations)
    except NoSetupTableError as e:
        logs.critical(e, exc_info=False)


@migration.command(cmd="down", help="Rollback the database to the specified migration")
async def migrate_down_main(
    conn: Connection = Derived(get_pg),
    sql_dir: Path = MigrationDir,
    nb_rollbacks: int | None = Option(None, "-n", "--nb-rollbacks", help="Number of rollbacks to apply"),
    migration_id: str | None = Option(None, "-m", "--migration-id", help="Migration id to rollback to"),
):
    from .migration import migrate_down, NoSetupTableError

    try:
        await migrate_down(conn, folder=sql_dir, migration_id=migration_id, nb_migrations=nb_rollbacks)
    except NoSetupTableError as e:
        logs.critical(e, exc_info=False)


@migration.command(cmd="setup", help="Create the tables that will contain the migration metadata")
async def migrate_setup_main(pg_conn: Connection = Derived(get_pg)):
    from .migration import migrate_setup

    await migrate_setup(pg_conn)


@migration.command(cmd="verify", help="Verify that a migration is valid")
def migrate_verify_main(
    db: str = PgDatabase,
    pg_host: str = PgHost,
    pg_port: int = PgPort,
    pg_user: str = PgUser,
    pg_password: Password = PgPassword,
    schemas: list[str] = Option(..., "--schemas", help="Schemas impacted by the migration"),
    sql_dir: Path = MigrationDir,
):
    os.environ["PG_HOST"] = pg_host
    os.environ["PG_PORT"] = str(pg_port)
    os.environ["PG_USER"] = pg_user
    os.environ["PG_PASSWORD"] = pg_password

    from .migration import migrate_verify, MigrationError

    try:
        with tempfile.TemporaryDirectory() as dump_dir:
            migrate_verify(
                database=db,
                migration_folder=sql_dir,
                schemas=schemas,
                dump_dir=Path(dump_dir),
            )
    except MigrationError as e:
        logs.error(e.msg)
        logs.debug(e.diff)


@migration.command(cmd="verify-files", help="Verify the files are correctly ordered")
def verify_files(
    sql_dir: Path = MigrationDir,
    no_raise_error: bool = Option(False, "--no-raise", help="Raise an error if the files are not correctly ordered"),
):
    from .utils import verify_migration_files

    has_errors = verify_migration_files(sql_dir, raise_error=not no_raise_error)
    if has_errors:
        raise CommandError("Files are not correctly ordered")
    else:
        CONSOLE.print("[green]Files are correctly ordered[/green]")


@migration.command(cmd="reorder-files", help="Reorder the files")
def reorder_files(
    migrations_dir: Path = MigrationDir,
    output_dir: Path | None = Option(
        None, "--output-dir", "-o", help="Output directory", raise_path_does_not_exist=False
    ),
    migration_ids: list[str] | None = Option(None, "--ids", "-l", help="Last migration ids (in descending order)"),
    reorder_by: list[Literal["last-applied", "last"]] | None = Option(
        None, "--by", help="Which method to use for the reorder (only useful if migration ids are speccified)"
    ),
    skip_verify: bool = Option(False, "--skip-verify", help="Skip verification after reordering"),
):
    from .reorder import reorder_files, reorder_files_by_applied_migrations, reorder_files_by_last
    from .utils import verify_migration_files, MigrationFileError

    folder = migrations_dir
    # Creating the output dir
    if output_dir is not None:
        if output_dir.exists():
            shutil.rmtree(output_dir)
        shutil.copytree(migrations_dir, output_dir)
        folder = output_dir

    match reorder_by:
        case "last-applied":
            reorder_files_by_applied_migrations(folder, last_applied_ids=migration_ids)
        case "last":
            reorder_files_by_last(folder, last_ids=migration_ids)
        case _:
            reorder_files(folder)

    if not skip_verify:
        try:
            verify_migration_files(folder)
        except MigrationFileError as e:
            raise CommandError(e.message)


@migration.command(cmd="verify-migrations", help="Verify that the migrations are applied correctly")
async def verify_migrations(
    pg_conn: Connection = Derived(get_pg),
    migrations_dir: Path = MigrationDir,
):
    from .migration import verify_migrations

    await verify_migrations(pg_conn, migrations_dir)



from .create_files import create_new_migration
from .migration import (
    migrate_down,
    migrate_up,
    migrate_verify,
    migrate_setup,
    verify_migrations,
    get_missing_migrations,
)
from .run import migration
from .utils import *
from .reorder import *



import datetime as dt
from pathlib import Path

from padmy.logs import logs
from .utils import get_files, utc_now, iter_migration_files, MigrationFile, verify_migration_files, MigrationErrorType

__all__ = ("reorder_files", "reorder_files_by_applied_migrations", "rename_files", "reorder_files_by_last")


def reorder_files_by_applied_migrations(folder: Path, last_applied_ids: list[str]):
    """
    Reorder the migration files given the last N commits (in descending order).
    For instance, let's say we have the following migration files:

            0000_file1.sql
            0001_file2.sql
            0002_file3.sql
            0003_file4.sql
            0004_file5.sql

    and we want to reorder the files given the last 2 commits: 0003 and 0001 that have already been applied,
    the new order will be:

            0000_file1.sql
            0001_file2.sql <= last commit -1
            0003_file4.sql <= last commit
            0002_file3.sql <= moved after the last commit
            0004_file5.sql <= moved after the last commit

    This is useful when you have migrations applied on one branch (let's say master) that are not
    in the same order as the branch you are currently working on (like develop).

    """
    if not last_applied_ids:
        logs.warning("No migration ids provided, skipping reorder")
        return

    _migration_ids = set(last_applied_ids)

    to_reorder_files, commit_files, after_files = [], [], []
    last_up_file, last_down_file = None, None
    files = get_files(folder, reverse=True)
    # Iterating files to find the one to reorder starting from the most recent migrations
    for up_file, down_file in iter_migration_files(files):
        # If we don't have any more migrations to reorder, we can stop
        if not _migration_ids:
            # Saving the last up and down files for the rewrite_headers function later
            last_up_file, last_down_file = up_file, down_file
            break

        # If we encounter a migration file in the last commits list, we add it to the commit_files list
        if up_file.file_id in _migration_ids:
            _migration_ids -= {up_file.file_id}
            commit_files.append((up_file, down_file))
            continue
        # If the size in unchanged, we did not meet any migration file to reorder
        if len(_migration_ids) == len(last_applied_ids):
            after_files.append((up_file, down_file))
        # Otherwise, we add it to the to_reorder_files list
        else:
            to_reorder_files.append((up_file, down_file))

    if _migration_ids:
        raise ValueError("Some migration ids were not found")

    new_order_files = commit_files + to_reorder_files + after_files

    # We rename the files to reflect the new order
    rename_files(utc_now(), last_migrations=(last_up_file, last_down_file), migration_files=new_order_files)


def _fix_file(
    files: tuple[MigrationFile, MigrationFile],
    prev_files: tuple[MigrationFile, MigrationFile],
    error_type: MigrationErrorType,
):
    up_file, down_file = files
    prev_up_file, prev_down_file = prev_files

    match error_type:
        case "header":
            if up_file.header is None or down_file.header is None:
                raise ValueError("Header is missing")
            up_file.header.prev_file, down_file.header.prev_file = prev_up_file.path.name, prev_down_file.path.name
            up_file.write_header()
            down_file.write_header()
        case _:
            raise NotImplementedError(f"Reorder error type {error_type} is not implemented")


def reorder_files(folder: Path):
    invalid_files = verify_migration_files(folder, raise_error=False)
    if not invalid_files:
        logs.info("All files are correctly ordered")
        return
    logs.info(f"Found {len(invalid_files)} files to reorder")

    file_errors = {error.file_id: error for _, _, error in invalid_files}

    prev_up_file, prev_down_file = None, None
    for up_file, down_file in iter_migration_files(get_files(folder)):
        # Skipping the first iteration
        if prev_up_file is None or prev_down_file is None:
            prev_up_file, prev_down_file = (up_file, down_file)
            continue

        _error = file_errors.get(up_file.file_id)
        if _error is not None:
            _fix_file((up_file, down_file), (prev_up_file, prev_down_file), _error.error_type)
        prev_up_file, prev_down_file = up_file, down_file

    logs.info("Done")


def rename_files(
    last_ts: dt.datetime,
    last_migrations: tuple[MigrationFile | None, MigrationFile | None],
    migration_files: list[tuple[MigrationFile, MigrationFile]],
):
    """
    Rename the migration files to reflect the new order.
    Also rewrite the headers accordingly.
    """
    _prev_up_file, _prev_down_file = last_migrations
    for i, (_up_file, _down_file) in enumerate(migration_files):
        _up_file.replace_ts(last_ts + dt.timedelta(seconds=i))
        _down_file.replace_ts(last_ts + dt.timedelta(seconds=i))
        # Rewrite headers
        if _prev_down_file is not None and _prev_up_file is not None:
            _fix_file((_up_file, _down_file), (_prev_up_file, _prev_down_file), "header")
        _prev_up_file, _prev_down_file = _up_file, _down_file


def reorder_files_by_last(folder: Path, last_ids: list[str]):
    """
    Reorder the migration files given the last N commits (in descending order).
    For instance, let's say we have the following migration files:

        0000_file1.sql
        0001_file2.sql
        0003_file3.sql
        0004_file4.sql
        0005_file5.sql

    And we want to reorder the files given the last 2 commits: 0004 and 003,
    the new order will be:

        0000_file1.sql
        0001_file2.sql
        0005_file5.sql
        0003_file3.sql <= moved to last -1
        0004_file4.sql <= moved to last
    """
    if not last_ids:
        logs.warning("No migration ids provided, skipping reorder")
        return
    _file_ids = set(last_ids)
    last_ts = utc_now()
    for up_file, down_file in iter_migration_files(get_files(folder)):
        if not _file_ids:
            break
        if up_file.file_id in _file_ids:
            up_file.replace_ts(last_ts)
            down_file.replace_ts(last_ts)
            last_ts += dt.timedelta(microseconds=1)
            _file_ids.remove(up_file.file_id)

    if _file_ids:
        raise ValueError("Some migration ids were not found")

    reorder_files(folder)



import logging
from rich.logging import RichHandler

logs = logging.getLogger("padmy")


def setup_logging(level: int):
    logging.basicConfig(
        datefmt="%H:%M:%S",
        format="%(message)s",
        handlers=[RichHandler(rich_tracebacks=False, show_path=False)],
    )
    logs.setLevel(level)



import datetime as dt
import sys
import uuid
from typing import Literal, get_args, TypeGuard

if sys.version_info >= (3, 12):
    UTC = dt.UTC
else:
    from pytz import UTC  # pyright: ignore[reportMissingModuleSource]

import dataclasses
import textwrap
from itertools import groupby
from operator import attrgetter
from pathlib import Path
from padmy.logs import logs

from padmy.utils import exec_cmd

__all__ = (
    "Header",
    "MigrationFile",
    "get_files",
    "iter_migration_files",
    "verify_migration_files",
    "get_git_email",
    "utc_now",
    "MigrationFileError",
    "MigrationErrorType",
    "parse_filename",
)


PREFIXES = {
    "author": "-- Author:",
    "prev-file": "-- Prev-file:",
    "version": "-- Version:",
    "skip-verify": "-- Skip-verify:",
}


def get_git_email():
    return exec_cmd(["git", "config", "user.email"]).strip()


@dataclasses.dataclass
class Header:
    prev_file: str | None
    author: str | None
    version: str | None
    skip_verify: bool = False
    skip_reason: str | None = None

    def __post_init__(self):
        if self.skip_verify and self.skip_reason is None:
            self.skip_reason = "no reason provided"

    @property
    def is_empty(self):
        return not any([self.prev_file, self.author, self.version])

    @classmethod
    def from_text(cls, text: str):
        prev_file = None
        author = None
        version = None
        skip_verify = False
        skip_reason = None
        for line in text.split("\n"):
            if line.startswith(PREFIXES["prev-file"]):
                prev_file = line.split(":")[1].strip()
            elif line.startswith(PREFIXES["author"]):
                author = line.split(":")[1].strip()
            elif line.startswith(PREFIXES["version"]):
                version = line.split(":")[1].strip()
            elif line.startswith(PREFIXES["skip-verify"]):
                skip_reason = line.split(":")[1].strip().lower()
                skip_verify = True
        return cls(prev_file, author, version, skip_verify=skip_verify, skip_reason=skip_reason)

    def as_text(self):
        _header = [
            f"{PREFIXES['prev-file']} {self.prev_file or ''}",
            f"{PREFIXES['author']} {self.author or ''}",
        ]
        if self.version is not None:
            _header.append(f"{PREFIXES['version']} {self.version}")
        if self.skip_verify:
            _header.append(f"{PREFIXES['skip-verify']} {self.skip_reason or 'no reason provided'}")

        file_header = textwrap.dedent("\n".join(_header)).strip()
        return file_header


@dataclasses.dataclass
class MigrationFile:
    ts: dt.datetime
    file_id: str
    file_type: str
    path: Path
    header: Header | None = None

    @property
    def name(self):
        return f"{int(self.ts.timestamp())}-{self.file_id}-{self.file_type}.sql"

    @property
    def skip_verify(self) -> bool:
        return self.header.skip_verify if self.header else False

    def replace_ts(self, ts: dt.datetime):
        self.ts = ts
        new_path = self.path.with_name(self.name)
        self.path = self.path.rename(new_path)

    def write_header(self):
        if not self.header:
            return
        # Remove the first lines starting with --
        if self.path.exists():
            lines = [
                _line for _line in self.path.read_text().split("\n") if not _line.startswith(tuple(PREFIXES.values()))
            ]
            _lines = "\n" + "\n".join(lines)
        else:
            _lines = "\n"
        new_text = self.header.as_text() + _lines
        self.path.write_text(new_text)

    @staticmethod
    def generate_base_name(ts: int | None = None, file_id: str | None = None) -> str:
        """
        Get a base name based on the timestamp and file_id
        to be used to generated file name later ({base_name}-{file_type}.sql)
        """
        _file_id = str(uuid.uuid4())[:8] if file_id is None else file_id
        _ts = ts if ts is not None else int(utc_now().timestamp())
        return f"{_ts}-{_file_id}"

    @classmethod
    def from_file(cls, path: Path):
        filename_infos = parse_filename(path.name)
        header = Header.from_text(path.read_text())
        return cls(
            ts=filename_infos["file_ts"],
            file_id=filename_infos["file_id"],
            file_type=filename_infos["migration_type"],
            path=path,
            header=header if not header.is_empty else None,
        )


def parse_filename(filename: str) -> dict:
    ts, file_id, file_type = filename.split("-")
    infos = {
        "file_ts": dt.datetime.fromtimestamp(int(ts), tz=UTC).replace(tzinfo=None),
        "file_id": file_id,
        "migration_type": file_type.replace(".sql", ""),
    }
    return infos


def get_files(folder: Path, reverse: bool = False, up_only: bool = False) -> list[MigrationFile]:
    """Returns the migration files in ascending order"""
    files = []
    pattern = "*.sql" if not up_only else "*-up.sql"
    for file in folder.glob(pattern):
        files.append(MigrationFile.from_file(file))

    return sorted(files, key=attrgetter("ts", "file_id"), reverse=reverse)


def iter_migration_files(files: list[MigrationFile]):
    for _file_id, _files_it in groupby(files, lambda x: x.file_id):
        _files: list[MigrationFile] = list(_files_it)

        _up_files = [x for x in _files if x.file_type == "up"]
        _down_files = [x for x in _files if x.file_type == "down"]

        if len(_up_files) != 1:
            raise ValueError(
                f'Found {len(_up_files)} "up" files (file_id: {_up_files[0].file_id})'
                if _up_files
                else "No up file found"
            )
        if len(_down_files) != 1:
            raise ValueError(
                f'Found {len(_down_files)} "down" files (file_id: {_down_files[0].file_id})'
                if _down_files
                else "No down file found"
            )

        yield _up_files[0], _down_files[0]


MigrationErrorType = Literal["order", "header", "duplicate"]


def is_migration_error_type(value: str) -> TypeGuard[MigrationErrorType]:
    return value in get_args(MigrationErrorType)


class MigrationFileError(Exception):
    def __init__(self, error_type: MigrationErrorType, message: str, file_id: str):
        self._error_type = error_type
        self.message = message
        self.file_id = file_id

    @property
    def error_type(self) -> MigrationErrorType:
        if is_migration_error_type(self._error_type):
            return self._error_type
        raise ValueError(f"Invalid error type {self._error_type!r}")


def verify_migration_files(
    migration_dir: Path, *, raise_error: bool = True
) -> list[tuple[MigrationFile, MigrationFile, MigrationFileError]]:
    """
    Verifies that the migration files are in order.
    If raise_error is set to False, it will return weather or not the files are in order
    and a list of files that are not correct (with the associated error message)
    """
    prev_files = None
    error_files = []
    _ids = set()
    for _file in iter_migration_files(get_files(migration_dir)):
        if prev_files is None:
            prev_files = _file
            continue
        prev_up, prev_down = prev_files
        up, down = _file

        if up.file_id in _ids:
            raise MigrationFileError("duplicate", f"Duplicate file_id: {up.file_id}", file_id=up.file_id)
        _ids.add(up.file_id)

        try:
            if prev_up.ts > up.ts:
                raise MigrationFileError(
                    "order", f"Files are not in order: {prev_up.path.name} > {up.path.name}", file_id=up.file_id
                )
            if prev_down.ts > down.ts:
                raise MigrationFileError(
                    "order", f"Files are not in order: {prev_down.path.name} > {down.path.name}", file_id=down.file_id
                )

            if up.header is not None and up.header.prev_file != prev_up.path.name:
                raise MigrationFileError(
                    "header",
                    f"Invalid header for up file {up.path.name} : "
                    f"(expected {prev_up.path.name!r} got {up.header.prev_file!r})",
                    file_id=up.file_id,
                )
            if down.header is not None and down.header.prev_file != prev_down.path.name:
                raise MigrationFileError(
                    "header",
                    f"Invalid header for down file {down.path.name} : "
                    f"(expected {prev_down.path.name!r} got {down.header.prev_file!r})",
                    file_id=down.file_id,
                )
        except MigrationFileError as e:
            if raise_error:
                raise e
            logs.warning(e.message)
            error_files.append((up, down, e))
        else:
            logs.debug(f"Valid header for file {up.path.name}")

        prev_files = _file

    return error_files


def utc_now():
    return dt.datetime.now(UTC).replace(tzinfo=None)
